# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import abc

from mako import parsetree
from mako.ast import PythonFragment

from .exceptions import ForbiddenMakoTemplateException


class MakoNodeCodeExtractor(object):
    @abc.abstractmethod
    def extract(self, node):
        """处理 Mako Lexer 分割出来的 code 对象，返回需要检测的 python 代码，返回 None 表示该节点不需要处理

        :param node: mako parsetree node
        :return: 需要处理的代码，或 None
        """
        raise NotImplementedError()


class StrictMakoNodeCodeExtractor(MakoNodeCodeExtractor):
    def extract(self, node):
        if isinstance(node, parsetree.Code) or isinstance(node, parsetree.Expression):
            return node.text
        elif isinstance(node, parsetree.ControlLine):
            if node.isend:
                return None
            return PythonFragment(node.text).code
        elif isinstance(node, parsetree.Text):
            return None
        else:
            raise ForbiddenMakoTemplateException("不支持[{}]节点".format(node.__class__.__name__))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.component_framework.constants import LEGACY_PLUGINS_VERSION
from pipeline.exceptions import ComponentNotExistException


class ComponentLibrary(object):
    components = {}

    def __new__(cls, *args, **kwargs):
        if args:
            component_code = args[0]
        else:
            component_code = kwargs.get("component_code", None)
        version = kwargs.get("version", None)
        if not component_code:
            raise ValueError(
                "please pass a component_code in args or kwargs: "
                "ComponentLibrary('code') or ComponentLibrary(component_code='code')"
            )
        return cls.get_component_class(component_code=component_code, version=version)

    @classmethod
    def component_list(cls):
        components = []
        for _, component_map in cls.components.items():
            components.extend(component_map.values())

        return components

    @classmethod
    def get_component_class(cls, component_code, version=None):
        version = version or LEGACY_PLUGINS_VERSION
        component_cls = cls.components.get(component_code, {}).get(version)
        if component_cls is None:
            raise ComponentNotExistException("component %s does not exist." % component_code)
        return component_cls

    @classmethod
    def get_component(cls, component_code, data_dict, version=None):
        version = version or LEGACY_PLUGINS_VERSION
        return cls.get_component_class(component_code=component_code, version=version)(data_dict)

    @classmethod
    def register_component(cls, component_code, version, component_cls):
        cls.components.setdefault(component_code, {})[version] = component_cls

    @classmethod
    def codes(cls):
        return cls.components.keys()

    @classmethod
    def versions(cls, code):
        return cls.components.get(code, {}).keys()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.28 on 2020-07-01 11:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0026_auto_20200610_1442"),
    ]

    operations = [
        migrations.CreateModel(
            name="SendFailedCeleryTask",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(max_length=1024, verbose_name="任务名")),
                ("kwargs", models.TextField(verbose_name="任务参数")),
                (
                    "type",
                    models.IntegerField(
                        choices=[
                            (0, "empty"),
                            (1, "process"),
                            (2, "node"),
                            (3, "schedule"),
                        ],
                        verbose_name="任务类型",
                    ),
                ),
                ("extra_kwargs", models.TextField(verbose_name="额外参数")),
                ("exec_trace", models.TextField(verbose_name="错误信息")),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="创建时间"),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-03-03 07:30
from __future__ import unicode_literals

from django.db import migrations, models
import pipeline.engine.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0024_auto_20200224_0308"),
    ]

    operations = [
        migrations.CreateModel(
            name="MultiCallbackData",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="自增ID")),
                ("schedule_id", models.CharField(max_length=64, verbose_name="回调服务ID")),
                ("data", pipeline.engine.models.fields.IOField(verbose_name="回调数据")),
            ],
        ),
    ]



# Generated by Django 2.2.19 on 2021-03-09 03:41

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CallbackData",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(max_length=33, verbose_name="节点 ID")),
                ("version", models.CharField(max_length=33, verbose_name="状态版本")),
                ("data", models.TextField(verbose_name="回调数据")),
            ],
        ),
        migrations.CreateModel(
            name="ContextOutputs",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("pipeline_id", models.CharField(max_length=33, unique=True, verbose_name="流程 ID")),
                ("outputs", models.TextField(verbose_name="输出配置")),
            ],
        ),
        migrations.CreateModel(
            name="Data",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(db_index=True, max_length=33, unique=True, verbose_name="节点 ID")),
                ("inputs", models.TextField(verbose_name="原始输入数据")),
                ("outputs", models.TextField(verbose_name="原始输出数据")),
            ],
        ),
        migrations.CreateModel(
            name="ExecutionData",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(db_index=True, max_length=33, unique=True, verbose_name="节点 ID")),
                ("inputs_serializer", models.CharField(max_length=32, verbose_name="输入序列化器")),
                ("outputs_serializer", models.CharField(max_length=32, verbose_name="输出序列化器")),
                ("inputs", models.TextField(verbose_name="节点执行输入数据")),
                ("outputs", models.TextField(verbose_name="节点执行输出数据")),
            ],
        ),
        migrations.CreateModel(
            name="Node",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(db_index=True, max_length=33, verbose_name="节点 ID")),
                ("detail", models.TextField(verbose_name="节点详情")),
            ],
        ),
        migrations.CreateModel(
            name="Process",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("parent_id", models.BigIntegerField(db_index=True, default=-1, verbose_name="父进程 ID")),
                ("ack_num", models.IntegerField(default=0, verbose_name="收到子进程 ACK 数量")),
                ("need_ack", models.IntegerField(default=-1, verbose_name="需要收到的子进程 ACK 数量")),
                ("asleep", models.BooleanField(default=True, verbose_name="是否处于休眠状态")),
                ("suspended", models.BooleanField(default=False, verbose_name="是否处于暂停状态")),
                ("frozen", models.BooleanField(default=False, verbose_name="是否处于冻结状态")),
                ("dead", models.BooleanField(default=False, verbose_name="是否已经死亡")),
                ("last_heartbeat", models.DateTimeField(auto_now_add=True, db_index=True, verbose_name="上次心跳时间")),
                ("destination_id", models.CharField(default="", max_length=33, verbose_name="执行终点 ID")),
                ("current_node_id", models.CharField(db_index=True, default="", max_length=33, verbose_name="当前节点 ID")),
                ("root_pipeline_id", models.CharField(max_length=33, verbose_name="根流程 ID")),
                (
                    "suspended_by",
                    models.CharField(db_index=True, default="", max_length=33, verbose_name="导致进程暂停的节点 ID"),
                ),
                ("priority", models.IntegerField(verbose_name="优先级")),
                ("queue", models.CharField(default="", max_length=128, verbose_name="所属队列")),
                ("pipeline_stack", models.TextField(default="[]", verbose_name="流程栈")),
            ],
        ),
        migrations.CreateModel(
            name="State",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(max_length=33, unique=True, verbose_name="节点 ID")),
                ("root_id", models.CharField(db_index=True, default="", max_length=33, verbose_name="根节点 ID")),
                ("parent_id", models.CharField(db_index=True, default="", max_length=33, verbose_name="父节点 ID")),
                ("name", models.CharField(max_length=64, verbose_name="状态名")),
                ("version", models.CharField(max_length=33, verbose_name="状态版本")),
                ("loop", models.IntegerField(default=1, verbose_name="循环次数")),
                ("retry", models.IntegerField(default=0, verbose_name="重试次数")),
                ("skip", models.BooleanField(default=False, verbose_name="是否跳过")),
                ("error_ignored", models.BooleanField(default=False, verbose_name="是否出错后自动忽略")),
                ("created_time", models.DateTimeField(auto_now_add=True, verbose_name="创建时间")),
                ("started_time", models.DateTimeField(null=True, verbose_name="开始时间")),
                ("archived_time", models.DateTimeField(null=True, verbose_name="归档时间")),
            ],
        ),
        migrations.CreateModel(
            name="Schedule",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("type", models.IntegerField(verbose_name="调度类型")),
                ("process_id", models.BigIntegerField(default=-1, verbose_name="进程 ID")),
                ("node_id", models.CharField(max_length=33, verbose_name="节点 ID")),
                ("finished", models.BooleanField(default=False, verbose_name="是否已完成")),
                ("expired", models.BooleanField(default=False, verbose_name="是否已过期")),
                ("scheduling", models.BooleanField(default=False, verbose_name="是否正在调度")),
                ("version", models.CharField(max_length=33, verbose_name="状态版本")),
                ("schedule_times", models.IntegerField(default=0, verbose_name="被调度次数")),
            ],
            options={"unique_together": {("node_id", "version")},},
        ),
        migrations.CreateModel(
            name="LogEntry",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(max_length=33, verbose_name="节点 ID")),
                ("loop", models.IntegerField(default=1, verbose_name="循环次数")),
                ("logger_name", models.CharField(max_length=128, verbose_name="logger 名称")),
                ("level_name", models.CharField(max_length=32, verbose_name="日志等级")),
                ("message", models.TextField(null=True, verbose_name="日志内容")),
                ("logged_at", models.DateTimeField(auto_now_add=True, db_index=True, verbose_name="输出时间")),
            ],
            options={"index_together": {("node_id", "loop")},},
        ),
        migrations.CreateModel(
            name="ExecutionHistory",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("node_id", models.CharField(max_length=33, verbose_name="节点 ID")),
                ("loop", models.IntegerField(default=1, verbose_name="循环次数")),
                ("retry", models.IntegerField(default=0, verbose_name="重试次数")),
                ("skip", models.BooleanField(default=False, verbose_name="是否跳过")),
                ("version", models.CharField(max_length=33, verbose_name="状态版本")),
                ("started_time", models.DateTimeField(verbose_name="开始时间")),
                ("archived_time", models.DateTimeField(verbose_name="归档时间")),
                ("inputs_serializer", models.CharField(max_length=32, verbose_name="输入序列化器")),
                ("outputs_serializer", models.CharField(max_length=32, verbose_name="输出序列化器")),
                ("inputs", models.TextField(verbose_name="节点执行输入数据")),
                ("outputs", models.TextField(verbose_name="节点执行输出数据")),
            ],
            options={"index_together": {("node_id", "loop")},},
        ),
        migrations.CreateModel(
            name="ContextValue",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False, verbose_name="ID")),
                ("pipeline_id", models.CharField(max_length=33, verbose_name="流程 ID")),
                ("key", models.CharField(max_length=128, verbose_name="变量 key")),
                ("type", models.IntegerField(verbose_name="变量类型")),
                ("serializer", models.CharField(max_length=32, verbose_name="序列化器")),
                ("code", models.CharField(default="", max_length=128, verbose_name="计算型变量类型唯一标志")),
                ("value", models.TextField(verbose_name="变量值")),
                ("references", models.TextField(verbose_name="所有对其他变量直接或间接的引用")),
            ],
            options={"unique_together": {("pipeline_id", "key")},},
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy

from pipeline.core.data.expression import ConstantTemplate, deformat_constant_key
from pipeline.exceptions import ConstantNotExistException, ConstantReferenceException
from pipeline.utils.graph import Graph


class ConstantPool(object):
    def __init__(self, pool, lazy=False):
        self.raw_pool = pool
        self.pool = None

        if not lazy:
            self.resolve()

    def resolve(self):
        if self.pool:
            return

        refs = self.get_reference_info()

        nodes = list(refs.keys())
        flows = []
        for node in nodes:
            for ref in refs[node]:
                if ref in nodes:
                    flows.append([node, ref])
        graph = Graph(nodes, flows)
        # circle reference check
        trace = graph.get_cycle()
        if trace:
            raise ConstantReferenceException("Exist circle reference between constants: %s" % "->".join(trace))

        # resolve the constants reference
        pool = {}
        temp_pool = copy.deepcopy(self.raw_pool)
        # get those constants which are referenced only(not refer other constants)
        referenced_only = ConstantPool._get_referenced_only(temp_pool)
        while temp_pool:
            for ref in referenced_only:
                value = temp_pool[ref]["value"]

                # resolve those constants which reference the 'ref'
                for key, info in list(temp_pool.items()):
                    maps = {deformat_constant_key(ref): value}
                    temp_pool[key]["value"] = ConstantTemplate(info["value"]).resolve_data(maps)

                pool[ref] = temp_pool[ref]
                temp_pool.pop(ref)
            referenced_only = ConstantPool._get_referenced_only(temp_pool)

        self.pool = pool

    @staticmethod
    def _get_referenced_only(pool):
        referenced_only = []
        for key, info in list(pool.items()):
            reference = ConstantTemplate(info["value"]).get_reference()
            formatted_reference = ["${%s}" % ref for ref in reference]
            reference = [c for c in formatted_reference if c in pool]
            if not reference:
                referenced_only.append(key)
        return referenced_only

    def get_reference_info(self, strict=True):
        refs = {}
        for key, info in list(self.raw_pool.items()):
            reference = ConstantTemplate(info["value"]).get_reference()
            formatted_reference = ["${%s}" % ref for ref in reference]
            ref = [c for c in formatted_reference if not strict or c in self.raw_pool]
            refs[key] = ref
        return refs

    def resolve_constant(self, constant):
        if not self.pool:
            self.resolve()

        if constant not in self.pool:
            raise ConstantNotExistException("constant %s not exist." % constant)
        return self.pool[constant]["value"]

    def resolve_value(self, val):
        if not self.pool:
            self.resolve()

        maps = {deformat_constant_key(key): self.pool[key]["value"] for key in self.pool}

        return ConstantTemplate(val).resolve_data(maps)



# Generated by Django 2.2.8 on 2020-06-10 06:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('engine', '0025_multicallbackdata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pipelineprocess',
            name='is_alive',
            field=models.BooleanField(db_index=True, default=True, verbose_name='该 process 是否还有效'),
        ),
        migrations.AlterField(
            model_name='pipelineprocess',
            name='is_frozen',
            field=models.BooleanField(db_index=True, default=False, verbose_name='该 process 是否被冻结'),
        ),
        migrations.AlterField(
            model_name='pipelineprocess',
            name='is_sleep',
            field=models.BooleanField(db_index=True, default=False, verbose_name='该 process 是否正在休眠'),
        ),
        migrations.AlterField(
            model_name='pipelineprocess',
            name='root_pipeline_id',
            field=models.CharField(db_index=True, max_length=32, verbose_name='根 pipeline 的 ID'),
        ),
        migrations.AlterField(
            model_name='scheduleservice',
            name='is_scheduling',
            field=models.BooleanField(db_index=True, default=False, verbose_name='是否正在被调度'),
        ),
        migrations.AlterField(
            model_name='noderelationship',
            name='distance',
            field=models.IntegerField(db_index=True, verbose_name='距离'),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import migrations

from pipeline.models import PipelineInstance, PipelineTemplate
from pipeline.contrib.statistics.models import InstanceInPipeline, TemplateInPipeline


def load_data(apps, schema_editor):
    # 清空数据
    TemplateInPipeline.objects.all().delete()
    InstanceInPipeline.objects.all().delete()
    template_list = PipelineTemplate.objects.filter(is_deleted=False)
    template_data = []
    for template in template_list:
        template_id = template.template_id
        try:
            result = statistics_total(template.data)
            data = TemplateInPipeline(
                template_id=template_id,
                atom_total=result["atom_total"],
                subprocess_total=result["subprocess_total"],
                gateways_total=result["gateways_total"],
            )
            template_data.append(data)
        except Exception:
            pass
    TemplateInPipeline.objects.bulk_create(template_data)

    instance_list = PipelineInstance.objects.filter(is_deleted=False)
    instance_data = []
    for instance in instance_list:
        instance_id = instance.instance_id
        try:
            result = statistics_total(instance.execution_data)
            data = InstanceInPipeline(
                instance_id=instance_id,
                atom_total=result["atom_total"],
                subprocess_total=result["subprocess_total"],
                gateways_total=result["gateways_total"],
            )
            instance_data.append(data)
        except Exception:
            pass
    InstanceInPipeline.objects.bulk_create(instance_data)


def statistics_total(pipeline_tree):
    atom_total = 0
    subprocess_total = 0
    tree_activities = pipeline_tree["activities"]
    # 获取网关数量
    gateways_total = len(pipeline_tree["gateways"])

    # 遍历activities节点
    for activity in tree_activities:
        activity_type = tree_activities[activity]["type"]
        if activity_type == "ServiceActivity":
            atom_total += 1
        elif activity_type == "SubProcess":
            subprocess_total += 1
    return {"atom_total": atom_total, "subprocess_total": subprocess_total, "gateways_total": gateways_total}


class Migration(migrations.Migration):
    dependencies = [("statistics", "0006_auto_20181115_1208")]
    operations = [migrations.RunPython(load_data)]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from typing import Optional

from bamboo_engine.eri import Service as ServiceInterface
from bamboo_engine.eri import Schedule, ExecutionData, CallbackData, ScheduleType

from pipeline.core.flow.activity import Service
from pipeline.core.data.base import DataObject
from pipeline.eri.log import get_logger


class ServiceWrapper(ServiceInterface):
    def __init__(self, service: Service):
        self.service = service

    def pre_execute(self, data: ExecutionData, root_pipeline_data: ExecutionData):
        """
        execute 执行前执行的逻辑

        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        """
        pre_execute = getattr(self.service, "pre_execute", None)
        if callable(pre_execute):
            return pre_execute(DataObject(inputs=data.inputs, outputs=data.outputs))

    def execute(self, data: ExecutionData, root_pipeline_data: ExecutionData) -> bool:
        """
        execute 逻辑

        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :return: 是否执行成功
        :rtype: bool
        """
        data_obj = DataObject(inputs=data.inputs, outputs=data.outputs)
        parent_data_obj = DataObject(inputs=root_pipeline_data.inputs, outputs=root_pipeline_data.outputs)

        try:
            execute_res = self.service.execute(data_obj, parent_data_obj)
        finally:
            # sync data object modification to execution data
            data.inputs = data_obj.inputs
            data.outputs = data_obj.outputs

        if execute_res is None:
            execute_res = True

        return execute_res

    def schedule(
        self,
        schedule: Schedule,
        data: ExecutionData,
        root_pipeline_data: ExecutionData,
        callback_data: Optional[CallbackData] = None,
    ) -> bool:
        """
        schedule 逻辑

        :param schedule: Schedule 对象
        :type schedule: Schedule
        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :param callback_data: 回调数据, defaults to None
        :type callback_data: Optional[CallbackData], optional
        :return: [description]
        :rtype: bool
        """
        data_obj = DataObject(inputs=data.inputs, outputs=data.outputs)
        parent_data_obj = DataObject(inputs=root_pipeline_data.inputs, outputs=root_pipeline_data.outputs)

        try:
            schedule_res = self.service.schedule(
                data_obj, parent_data_obj, callback_data.data if callback_data else None
            )
        except Exception as e:
            raise e
        finally:
            # sync data object modification to execution data
            data.inputs = data_obj.inputs
            data.outputs = data_obj.outputs

        if schedule_res is None:
            schedule_res = True

        return schedule_res

    def need_schedule(self) -> bool:
        """
        服务是否需要调度

        :return: 是否需要调度
        :rtype: bool
        """
        return self.service.need_schedule()

    def schedule_type(self) -> Optional[ScheduleType]:
        """
        服务调度类型

        :return: 调度类型
        :rtype: Optional[ScheduleType]
        """
        if not self.service.need_schedule():
            return None

        if self.service.interval:
            return ScheduleType.POLL

        if not self.service.multi_callback_enabled():
            return ScheduleType.CALLBACK

        return ScheduleType.MULTIPLE_CALLBACK

    def is_schedule_done(self) -> bool:
        """
        调度是否完成

        :return: 调度是否完成
        :rtype: bool
        """
        return self.service.is_schedule_finished()

    def schedule_after(
        self, schedule: Optional[Schedule], data: ExecutionData, root_pipeline_data: ExecutionData
    ) -> int:
        """
        计算下一次调度间隔

        :param schedule: 调度对象，未进行调度时传入为空
        :type schedule: Optional[Schedule]
        :param data: 节点执行数据
        :type data: ExecutionData
        :param root_pipeline_data: 根流程执行数据
        :type root_pipeline_data: ExecutionData
        :return: 调度间隔，单位为秒
        :rtype: int
        """
        if self.service.interval is None:
            return -1

        if schedule is None:
            return self.service.interval.next()

        # count will add in next, so minus 1 at here
        self.service.interval.count = schedule.times - 1

        return self.service.interval.next()

    def setup_runtime_attributes(self, **attrs):
        """
        装载运行时属性

        :param attrs: 运行时属性
        :type attrs: Dict[str, Any]
        """

        attrs["logger"] = get_logger(node_id=attrs["id"], loop=attrs["loop"], version=attrs["version"])
        self.service.setup_runtime_attrs(**attrs)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import importlib
import sys
import traceback
from contextlib import contextmanager

from mock import MagicMock, call, patch

from pipeline.core.data.base import DataObject
from pipeline.core.flow.io import SimpleItemSchema, ArrayItemSchema, ObjectItemSchema, ItemSchema
from pipeline.utils.uniqid import uniqid


@contextmanager
def patch_context(patchers):
    for patcher in patchers:
        patcher.start()

    yield

    for patcher in patchers:
        patcher.stop()


class ComponentTestMixin(object):
    def component_cls(self):
        raise NotImplementedError()

    def cases(self):
        raise NotImplementedError()

    def input_output_format_valid(self):
        component = self._component_cls({})
        bound_service = component.service()
        inputs_format = bound_service.inputs()
        self._format_valid(inputs_format, ["name", "key", "type", "schema", "required"])
        outputs_format = bound_service.outputs()
        self._format_valid(outputs_format, ["name", "key", "type", "schema"])

    @property
    def _cases(self):
        return self.cases()

    @property
    def _component_cls(self):
        return self.component_cls()

    @property
    def _component_cls_name(self):
        return self._component_cls.__name__

    @property
    def _failed_cases(self):
        return getattr(self, "__failed_cases", None)

    def _format_valid(self, component_format, format_keys):
        assert isinstance(component_format, list)
        for item in component_format:
            assert set(item.as_dict().keys()) == set(
                format_keys
            ), "item {} is expected to contain attributes {} but {} obtained".format(
                item.key, str(format_keys), str(item.as_dict().keys())
            )
            if item.schema is not None:
                assert item.type == item.schema.type, "type of {} is expected to be {} but {} obtained".format(
                    item.key, item.schema.type, item.type
                )
                self._item_schema_valid(item.schema)

    def _item_schema_valid(self, item_schema):
        common_keys = {"type", "description", "enum"}
        assert common_keys.issubset(
            set(item_schema.as_dict().keys())
        ), "ItemSchema should contain attributes type, description and enum"

        if isinstance(item_schema, SimpleItemSchema):
            return
        if isinstance(item_schema, ArrayItemSchema):
            assert hasattr(item_schema, "item_schema") and isinstance(
                item_schema.item_schema, ItemSchema
            ), "ArrayItemSchema should contain attribute item_schema"
            self._item_schema_valid(item_schema.item_schema)
            return
        if isinstance(item_schema, ObjectItemSchema):
            assert hasattr(item_schema, "property_schemas") and isinstance(
                item_schema.property_schemas, dict
            ), "ObjectItemSchema should contain attribute property_schemas with type dict"
            for child_item_schema in item_schema.property_schemas.values():
                self._item_schema_valid(child_item_schema)
            return
        raise AssertionError("item_schema type error: {}".format(item_schema.description))

    def _format_failure_message(self, no, name, msg):
        return "[{component_cls} case {no}] - [{name}] fail: {msg}".format(
            component_cls=self._component_cls_name, no=no + 1, name=name, msg=msg
        )

    def _do_case_assert(self, service, method, assertion, no, name, args=None, kwargs=None):

        args = args or [service]
        kwargs = kwargs or {}

        data = kwargs.get("data") or args[0]

        result = getattr(service, method)(*args, **kwargs)

        assert_success = result in [None, True]  # return none will consider as success
        do_continue = not assert_success

        assert_method = "assertTrue" if assert_success else "assertFalse"

        getattr(self, assert_method)(
            assertion.success,
            msg=self._format_failure_message(
                no=no,
                name=name,
                msg="{method} success assertion failed, {method} execute success".format(method=method),
            ),
        )

        self.assertDictEqual(
            data.outputs,
            assertion.outputs,
            msg=self._format_failure_message(
                no=no,
                name=name,
                msg="{method} outputs assertion failed,\nexcept: {e}\nactual: {a}".format(
                    method=method, e=assertion.outputs, a=data.outputs
                ),
            ),
        )

        return do_continue

    def _do_call_assertion(self, name, no, assertion):
        try:
            assertion.do_assert()
        except AssertionError as e:
            self.assertTrue(
                False,
                msg=self._format_failure_message(
                    no=no, name=name, msg="{func} call assert failed: {e}".format(func=assertion.func, e=e)
                ),
            )

    def _case_pass(self, case):
        sys.stdout.write(
            "\n[√] <{component}> - [{case_name}]\n".format(
                component=self._component_cls_name,
                case_name=case.name,
            )
        )

    def _case_fail(self, case):
        sys.stdout.write(
            "\n[×] <{component}> - [{case_name}]\n".format(
                component=self._component_cls_name,
                case_name=case.name,
            )
        )

        if not hasattr(self, "__failed_cases"):
            setattr(self, "__failed_cases", [])

        self._failed_cases.append(case)

    def _test_fail(self):
        raise AssertionError("{} cases fail".format([case.name for case in self._failed_cases]))

    def test_component(self):
        self.input_output_format_valid()

        component = self._component_cls({})

        for no, case in enumerate(self._cases):
            try:

                patchers = [patcher.mock_patcher() for patcher in case.patchers]

                with patch_context(patchers):

                    bound_service = component.service()

                    setattr(bound_service, "root_pipeline_id", case.root_pipeline_id)
                    setattr(bound_service, "id", case.service_id)
                    setattr(bound_service, "logger", MagicMock())

                    data = DataObject(inputs=case.inputs)
                    parent_data = DataObject(inputs=case.parent_data)

                    # execute result check
                    do_continue = self._do_case_assert(
                        service=bound_service,
                        method="execute",
                        args=(data, parent_data),
                        assertion=case.execute_assertion,
                        no=no,
                        name=case.name,
                    )

                    for call_assertion in case.execute_call_assertion:
                        self._do_call_assertion(name=case.name, no=no, assertion=call_assertion)

                    if do_continue:
                        self._case_pass(case)
                        continue

                    if bound_service.need_schedule():

                        if bound_service.interval is None:
                            # callback case
                            self._do_case_assert(
                                service=bound_service,
                                method="schedule",
                                args=(data, parent_data, case.schedule_assertion.callback_data),
                                assertion=case.schedule_assertion,
                                no=no,
                                name=case.name,
                            )

                        else:
                            # schedule case
                            assertions = case.schedule_assertion
                            assertions = assertions if isinstance(assertions, list) else [assertions]

                            for assertion in assertions:
                                do_continue = self._do_case_assert(
                                    service=bound_service,
                                    method="schedule",
                                    args=(data, parent_data),
                                    assertion=assertion,
                                    no=no,
                                    name=case.name,
                                )

                                self.assertEqual(
                                    assertion.schedule_finished,
                                    bound_service.is_schedule_finished(),
                                    msg=self._format_failure_message(
                                        no=no,
                                        name=case.name,
                                        msg="schedule_finished assertion failed:"
                                        "\nexpected: {expected}\nactual: {actual}".format(
                                            expected=assertion.schedule_finished,  # noqa
                                            actual=bound_service.is_schedule_finished(),
                                        ),
                                    ),
                                )  # noqa

                                if do_continue:
                                    break

                        for call_assertion in case.schedule_call_assertion:
                            self._do_call_assertion(name=case.name, no=no, assertion=call_assertion)

                    self._case_pass(case)

            except Exception:
                self._case_fail(case)
                sys.stdout.write("{}\n".format(traceback.format_exc()))

        if self._failed_cases:
            self._test_fail()


class ComponentTestCase(object):
    def __init__(
        self,
        inputs,
        parent_data,
        execute_assertion,
        schedule_assertion,
        name="",
        patchers=None,
        execute_call_assertion=None,
        schedule_call_assertion=None,
        service_id=None,
        root_pipeline_id=None,
    ):
        self.inputs = inputs
        self.parent_data = parent_data
        self.execute_assertion = execute_assertion
        self.execute_call_assertion = execute_call_assertion or []
        self.schedule_call_assertion = schedule_call_assertion or []
        self.schedule_assertion = schedule_assertion
        self.name = name
        self.patchers = patchers or []
        self.service_id = service_id or uniqid()
        self.root_pipeline_id = root_pipeline_id or uniqid()


class CallAssertion(object):
    def __init__(self, func, calls, any_order=False):
        self.func = func
        self.calls = calls
        self.any_order = any_order

    def do_assert(self):
        if not callable(self.func):
            module_and_func = self.func.rsplit(".", 1)
            mod_path = module_and_func[0]
            func_name = module_and_func[1]
            mod = importlib.import_module(mod_path)
            func = getattr(mod, func_name)
        else:
            func = self.func

        if not self.calls:
            func.assert_not_called()
        else:
            assert func.call_count == len(
                self.calls
            ), "Expected 'mock' have been called {expect} times. " "Called {actual} times".format(
                expect=len(self.calls), actual=func.call_count
            )
            func.assert_has_calls(calls=self.calls, any_order=self.any_order)

        func.reset_mock()


class Assertion(object):
    def __init__(self, success, outputs):
        self.success = success
        self.outputs = outputs


class ExecuteAssertion(Assertion):
    pass


class ScheduleAssertion(Assertion):
    def __init__(self, callback_data=None, schedule_finished=False, *args, **kwargs):
        self.callback_data = callback_data
        self.schedule_finished = schedule_finished
        super(ScheduleAssertion, self).__init__(*args, **kwargs)


class Patcher(object):
    def __init__(self, target, return_value=None, side_effect=None):
        self.target = target
        self.return_value = return_value
        self.side_effect = side_effect

    def mock_patcher(self):
        return patch(target=self.target, new=MagicMock(return_value=self.return_value, side_effect=self.side_effect))


Call = call



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from typing import Optional

from celery import task

from bamboo_engine import states
from bamboo_engine.engine import Engine

from pipeline.eri.runtime import BambooDjangoRuntime


@task(ignore_result=True)
def execute(process_id: int, node_id: str, root_pipeline_id: str = None, parent_pipeline_id: str = None):
    runtime = BambooDjangoRuntime()
    Engine(runtime).execute(
        process_id=process_id, node_id=node_id, root_pipeline_id=root_pipeline_id, parent_pipeline_id=parent_pipeline_id
    )


@task(ignore_result=True)
def schedule(process_id: int, node_id: str, schedule_id: str, callback_data_id: Optional[int]):
    runtime = BambooDjangoRuntime()
    Engine(runtime).schedule(
        process_id=process_id, node_id=node_id, schedule_id=schedule_id, callback_data_id=callback_data_id
    )


@task(ignore_result=True)
def timeout_check(self, process_id: int, node_id: str, version: str):
    runtime = BambooDjangoRuntime()
    state = runtime.get_state(node_id=node_id)
    if state.name == states.RUNNING and state.version == version:
        Engine(runtime).forced_fail_activity(node_id=node_id, ex_data="timeout kill")



# Generated by Django 2.2.16 on 2021-07-12 12:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eri", "0003_logentry_version"),
    ]

    operations = [
        migrations.AddField(
            model_name="state", name="inner_loop", field=models.IntegerField(default=1, verbose_name="子流程内部循环次数"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from pipeline.conf import default_settings
from pipeline.core.data.hydration import hydrate_node_data
from pipeline.core.flow.activity import ServiceActivity
from pipeline.django_signal_valve import valve
from pipeline.engine import signals
from pipeline.engine.models import Data, ScheduleService, Status

from .base import FlowElementHandler

logger = logging.getLogger("pipeline_engine")

__all__ = ["ServiceActivityHandler"]


class ServiceActivityHandler(FlowElementHandler):
    @staticmethod
    def element_cls():
        return ServiceActivity

    def handle(self, process, element, status):
        pre_execute_success = False
        success = False
        exception_occurred = False
        monitoring = False
        version = status.version
        root_pipeline = process.root_pipeline

        # rerun mode
        if status.loop > 1 and not element.on_retry():
            element.prepare_rerun_data()
            process.top_pipeline.context.recover_variable()

        elif element.on_retry():
            element.retry_at_current_exec()

        # set loop to data
        element.data.inputs._loop = status.loop + default_settings.PIPELINE_RERUN_INDEX_OFFSET
        element.data.outputs._loop = status.loop + default_settings.PIPELINE_RERUN_INDEX_OFFSET

        # pre output extract
        process.top_pipeline.context.extract_output(element, set_miss=False)

        # hydrate inputs
        hydrate_node_data(element)

        if element.timeout:
            logger.info("node {} {} start timeout monitor, timeout: {}".format(element.id, version, element.timeout))
            signals.service_activity_timeout_monitor_start.send(
                sender=element.__class__,
                node_id=element.id,
                version=version,
                root_pipeline_id=root_pipeline.id,
                countdown=element.timeout,
            )
            monitoring = True

        element.setup_runtime_attrs(
            id=element.id, root_pipeline_id=root_pipeline.id,
        )

        # pre_process inputs and execute service
        try:
            pre_execute_success = element.execute_pre_process(root_pipeline.data)
            if pre_execute_success:
                success = element.execute(root_pipeline.data)
        except Exception:
            if element.error_ignorable:
                # ignore exception
                pre_execute_success = True
                success = True
                exception_occurred = True
                element.ignore_error()
            ex_data = traceback.format_exc()
            element.data.outputs.ex_data = ex_data
            logger.error(ex_data)

        # process result
        if pre_execute_success is False or success is False:
            ex_data = element.data.get_one_of_outputs("ex_data")
            Status.objects.fail(element, ex_data)
            try:
                element.failure_handler(root_pipeline.data)
            except Exception:
                logger.error("failure_handler({}) failed: {}".format(element.id, traceback.format_exc()))

            if monitoring:
                signals.service_activity_timeout_monitor_end.send(
                    sender=element.__class__, node_id=element.id, version=version
                )
                logger.info("node {} {} timeout monitor revoke".format(element.id, version))

            # send activity error signal
            valve.send(
                signals,
                "activity_failed",
                sender=root_pipeline,
                pipeline_id=root_pipeline.id,
                pipeline_activity_id=element.id,
                subprocess_id_stack=process.subprocess_stack,
            )

            return self.HandleResult(next_node=None, should_return=False, should_sleep=True)
        else:
            is_error_ignored = element.error_ignorable and not element.get_result_bit()
            if element.need_schedule() and not exception_occurred and not is_error_ignored:
                # write data before schedule
                Data.objects.write_node_data(element)
                return self.HandleResult(
                    next_node=None,
                    should_return=True,
                    should_sleep=True,
                    after_sleep_call=ScheduleService.objects.set_schedule,
                    args=[],
                    kwargs=dict(
                        activity_id=element.id,
                        service_act=element.shell(),
                        process_id=process.id,
                        version=version,
                        parent_data=process.top_pipeline.data,
                    ),
                )

            process.top_pipeline.context.extract_output(element)
            error_ignorable = not element.get_result_bit()

            if monitoring:
                signals.service_activity_timeout_monitor_end.send(
                    sender=element.__class__, node_id=element.id, version=version
                )
                logger.info("node {} {} timeout monitor revoke".format(element.id, version))

            if not Status.objects.finish(element, error_ignorable):
                # has been forced failed
                return self.HandleResult(next_node=None, should_return=False, should_sleep=True)
            return self.HandleResult(next_node=element.next(), should_return=False, should_sleep=False)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import os

from django.core.management.base import BaseCommand
from django.template.loader import render_to_string

from pipeline.conf import settings


class Command(BaseCommand):
    help = "Generate Redis & Supervisor configuration file for pipeline"

    configs = {
        os.path.join(settings.BASE_DIR, "etc/redis.conf"): "redis/redis.tmpl",
        os.path.join(settings.BASE_DIR, "etc/supervisord.conf"): "supervisor/supervisor.tmpl",
    }

    var_paths = [os.path.join(settings.BASE_DIR, "var/log/"), os.path.join(settings.BASE_DIR, "var/run/")]

    def add_arguments(self, parser):
        parser.add_argument("-pc", dest="p_worker_num", default=2, help="Set number of worker bind with pipeline")
        parser.add_argument("-sc", dest="s_worker_num", default=2, help="Set the number of worker bind with schedule")
        parser.add_argument(
            "--worker",
            action="store_true",
            dest="is_worker",
            default=False,
            help="is worker process group (default False)",
        )
        parser.add_argument(
            "--master",
            action="store_true",
            dest="is_master",
            default=False,
            help="is master process group (default False)",
        )

    def handle(self, *args, **options):
        context = {
            "settings": settings,
            "is_master": options["is_master"],
            "is_worker": options["is_worker"],
            "p_worker_num": options["p_worker_num"],
            "s_worker_num": options["s_worker_num"],
            "uid": os.getuid(),
        }

        for path in self.var_paths:
            if not os.path.exists(path):
                os.makedirs(path)

        for target_path, template_name in list(self.configs.items()):
            dirname = os.path.dirname(target_path)
            if not os.path.exists(dirname):
                try:
                    os.makedirs(dirname)
                except Exception:
                    pass

            with open(target_path, "wb+") as f:
                f.write(render_to_string(template_name, context))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
from collections import namedtuple
from datetime import datetime

import pytz
from celery import schedules
from celery.utils.time import is_naive, make_aware

schedstate = namedtuple("schedstate", ("is_due", "next"))
logger = logging.getLogger("celery")


class TzAwareCrontab(schedules.crontab):
    """Timezone Aware Crontab."""

    def __init__(
        self, minute="*", hour="*", day_of_week="*", day_of_month="*", month_of_year="*", tz=pytz.utc, app=None,
    ):
        """Overwrite Crontab constructor to include a timezone argument."""
        self.tz = tz

        nowfun = self.nowfunc

        super(TzAwareCrontab, self).__init__(
            minute=minute,
            hour=hour,
            day_of_week=day_of_week,
            day_of_month=day_of_month,
            month_of_year=month_of_year,
            nowfun=nowfun,
            app=app,
        )

    def nowfunc(self):
        return self.tz.normalize(pytz.utc.localize(datetime.utcnow()))

    def is_due(self, last_run_at):
        """Calculate when the next run will take place.
        Return tuple of (is_due, next_time_to_check).
        The last_run_at argument needs to be timezone aware.
        """
        logger.debug("################### is_due begin ###################")
        logger.debug("native last_run_at: %s" % last_run_at)

        last_run_at = last_run_at.astimezone(self.tz)

        now = datetime.now(self.tz)
        logger.debug("last_run_at: %s" % last_run_at)
        logger.debug("now: %s" % now)

        rem_delta = self.remaining_estimate(last_run_at)

        logger.debug("rem_delta: %s" % rem_delta)
        logger.debug("next run at: %s" % (now + rem_delta))

        rem = max(rem_delta.total_seconds(), 0)
        due = rem == 0
        if due:
            rem_delta = self.remaining_estimate(self.now())
            rem = max(rem_delta.total_seconds(), 0)

        logger.debug("self: %s" % self)
        logger.debug("due: {} {} {}".format(self.tz, due, rem))
        logger.debug("################### is_due end ###################")
        return schedstate(due, rem)

    # Needed to support pickling
    def __repr__(self):
        return (
            "<crontab: {0._orig_minute} {0._orig_hour} {0._orig_day_of_week} {0._orig_day_of_month} "
            "{0._orig_month_of_year} (m/h/d/dM/MY), {0.tz}>".format(self)
        )

    def __reduce__(self):
        return (
            self.__class__,
            (
                self._orig_minute,
                self._orig_hour,
                self._orig_day_of_week,
                self._orig_day_of_month,
                self._orig_month_of_year,
                self.tz,
            ),
            None,
        )

    def __eq__(self, other):
        if isinstance(other, schedules.crontab):
            return (
                other.month_of_year == self.month_of_year
                and other.day_of_month == self.day_of_month
                and other.day_of_week == self.day_of_week
                and other.hour == self.hour
                and other.minute == self.minute
                and other.tz == self.tz
            )
        return NotImplemented

    def maybe_make_aware(self, dt):
        if not is_naive(dt):
            return dt
        return make_aware(dt, self.tz)

    def to_local(self, dt):
        return self.maybe_make_aware(dt)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from typing import List
from datetime import datetime

from bamboo_engine.eri import ExecutionHistory, ExecutionShortHistory

from pipeline.eri.models import ExecutionHistory as DBExecutionHistory
from pipeline.eri.imp.serializer import SerializerMixin


class ExecutionHistoryMixin(SerializerMixin):
    def add_history(
        self,
        node_id: str,
        started_time: datetime,
        archived_time: datetime,
        loop: int,
        skip: bool,
        retry: int,
        version: str,
        inputs: dict,
        outputs: dict,
    ) -> int:
        """
        为某个节点记录一次执行历史

        : param node_id: 节点 ID
        : type node_id: str
        : param started_time: 开始时间
        : type started_time: datetime
        : param archived_time: 归档时间
        : type archived_time: datetime
        : param loop: 重入计数
        : type loop: int
        : param skip: 是否跳过
        : type skip: bool
        : param retry: 重试次数
        : type retry: int
        : param version: 节点执行版本号
        : type version: str
        : param inputs: 输入数据
        : type inputs: dict
        : param outputs: 输出数据
        : type outputs: dict
        """
        inputs, inputs_serializer = self._serialize(inputs)
        outputs, outputs_serializer = self._serialize(outputs)
        return DBExecutionHistory.objects.create(
            node_id=node_id,
            loop=loop,
            retry=retry,
            skip=skip,
            version=version,
            started_time=started_time,
            archived_time=archived_time,
            inputs=inputs,
            inputs_serializer=inputs_serializer,
            outputs=outputs,
            outputs_serializer=outputs_serializer,
        ).id

    def get_histories(self, node_id: str, loop: int = -1) -> List[ExecutionHistory]:
        """
        返回某个节点的历史记录

        :param node_id: 节点 ID
        :type node_id: str
        :param loop: 重入次数, -1 表示不过滤重入次数
        :type loop: int, optional
        :return: 历史记录列表
        :rtype: List[History]
        """
        fields = {"node_id": node_id}
        if loop != -1:
            fields["loop"] = loop
        qs = DBExecutionHistory.objects.filter(**fields)

        return [
            ExecutionHistory(
                id=model.id,
                node_id=model.node_id,
                started_time=model.started_time,
                archived_time=model.archived_time,
                loop=model.loop,
                skip=model.skip,
                retry=model.retry,
                version=model.version,
                inputs=self._deserialize(model.inputs, model.inputs_serializer),
                outputs=self._deserialize(model.outputs, model.outputs_serializer),
            )
            for model in qs
        ]

    def get_short_histories(self, node_id: str, loop: int = -1) -> List[ExecutionShortHistory]:
        """
        返回某个节点的简要历史记录

        :param node_id: 节点 ID
        :type node_id: str
        :param loop: 重入次数, -1 表示不过滤重入次数
        :type loop: int, optional
        :return: 历史记录列表
        :rtype: List[ExecutionShortHistory]
        """
        fields = {"node_id": node_id}
        if loop != -1:
            fields["loop"] = loop
        qs = DBExecutionHistory.objects.filter(**fields).defer(
            "inputs", "inputs_serializer", "outputs", "outputs_serializer"
        )

        return [
            ExecutionShortHistory(
                id=model.id,
                node_id=model.node_id,
                started_time=model.started_time,
                archived_time=model.archived_time,
                loop=model.loop,
                skip=model.skip,
                retry=model.retry,
                version=model.version,
            )
            for model in qs
        ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-17 08:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("statistics", "0010_auto_20190304_1747"),
    ]

    operations = [
        migrations.AddField(
            model_name="componentexecutedata",
            name="version",
            field=models.CharField(default="legacy", max_length=255, verbose_name="插件版本"),
        ),
        migrations.AddField(
            model_name="componentintemplate",
            name="version",
            field=models.CharField(default="legacy", max_length=255, verbose_name="插件版本"),
        ),
        migrations.AlterField(
            model_name="componentexecutedata",
            name="subprocess_stack",
            field=models.TextField(default="[]", help_text="JSON 格式的列表", verbose_name="子流程堆栈"),
        ),
        migrations.AlterField(
            model_name="componentintemplate",
            name="subprocess_stack",
            field=models.TextField(default="[]", help_text="JSON 格式的列表", verbose_name="子流程堆栈"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from copy import deepcopy

from pipeline import exceptions
from pipeline.core.constants import PE


def format_to_list(notype):
    """
    format a data to list
    :return:
    """
    if isinstance(notype, list):
        return notype
    if not notype:
        return []
    return [notype]


def format_node_io_to_list(node, i=True, o=True):
    if i:
        node["incoming"] = format_to_list(node["incoming"])

    if o:
        node["outgoing"] = format_to_list(node["outgoing"])


def format_pipeline_tree_io_to_list(pipeline_tree):
    """
    :summary: format incoming and outgoing to list
    :param pipeline_tree:
    :return:
    """
    for act in list(pipeline_tree[PE.activities].values()):
        format_node_io_to_list(act, o=False)

    for gateway in list(pipeline_tree[PE.gateways].values()):
        format_node_io_to_list(gateway, o=False)

    format_node_io_to_list(pipeline_tree[PE.end_event], o=False)


def get_node_for_sequence(sid, tree, node_type):
    target_id = tree[PE.flows][sid][node_type]

    if target_id in tree[PE.activities]:
        return tree[PE.activities][target_id]
    elif target_id in tree[PE.gateways]:
        return tree[PE.gateways][target_id]
    elif target_id == tree[PE.end_event][PE.id]:
        return tree[PE.end_event]
    elif target_id == tree[PE.start_event][PE.id]:
        return tree[PE.start_event]

    raise exceptions.InvalidOperationException("node(%s) not in data" % target_id)


def get_nodes_dict(data):
    """
    get all FlowNodes of a pipeline
    """
    data = deepcopy(data)
    start = data[PE.start_event][PE.id]
    end = data[PE.end_event][PE.id]

    nodes = {start: data[PE.start_event], end: data[PE.end_event]}

    nodes.update(data[PE.activities])
    nodes.update(data[PE.gateways])

    for node in list(nodes.values()):
        # format to list
        node[PE.incoming] = format_to_list(node[PE.incoming])
        node[PE.outgoing] = format_to_list(node[PE.outgoing])

        node[PE.source] = [data[PE.flows][incoming][PE.source] for incoming in node[PE.incoming]]
        node[PE.target] = [data[PE.flows][outgoing][PE.target] for outgoing in node[PE.outgoing]]

    return nodes



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import contextlib
import logging
import traceback

from pipeline.conf import settings as pipeline_settings
from pipeline.core.flow.activity import SubProcess
from pipeline.engine import states
from pipeline.engine.core import context
from pipeline.engine.core.handlers import HandlersFactory
from pipeline.engine.models import NAME_MAX_LENGTH, FunctionSwitch, NodeRelationship, Status

logger = logging.getLogger("pipeline_engine")
celery_logger = logging.getLogger("celery")

RERUN_MAX_LIMIT = pipeline_settings.PIPELINE_RERUN_MAX_TIMES


@contextlib.contextmanager
def runtime_exception_handler(process):
    try:
        yield
    except Exception as e:
        logger.error(traceback.format_exc())
        process.exit_gracefully(e)


def run_loop(process):
    """
    pipeline 推进主循环
    :param process: 当前进程
    :return:
    """
    with runtime_exception_handler(process):
        while True:

            current_node = process.top_pipeline.node(process.current_node_id)
            celery_logger.info(
                "[pipeline-trace](root_pipeline: %s) execute node %s" % (process.root_pipeline_id, current_node.id)
            )

            # check child process destination
            if process.destination_id == current_node.id:
                try:
                    process.destroy_and_wake_up_parent(current_node.id)
                except Exception:
                    logger.error(traceback.format_exc())
                logger.info("child process(%s) finish." % process.id)
                return

            # check root pipeline status
            need_sleep, pipeline_state = process.root_sleep_check()
            if need_sleep:
                logger.info("pipeline(%s) turn to sleep." % process.root_pipeline.id)
                process.sleep(do_not_save=(pipeline_state == states.REVOKED))
                return

            # check subprocess status
            need_sleep, subproc_above = process.subproc_sleep_check()
            if need_sleep:
                logger.info("process(%s) turn to sleep." % process.root_pipeline.id)
                process.sleep(adjust_status=True, adjust_scope=subproc_above)
                return

            # check engine status
            if FunctionSwitch.objects.is_frozen():
                logger.info("pipeline(%s) have been frozen." % process.id)
                process.freeze()
                return

                # try to transit current node to running state
            name = (current_node.name or str(current_node.__class__))[:NAME_MAX_LENGTH]
            action = Status.objects.transit(id=current_node.id, to_state=states.RUNNING, start=True, name=name)

            # check rerun limit
            if (
                not isinstance(current_node, SubProcess)
                and RERUN_MAX_LIMIT != 0
                and action.extra.loop > RERUN_MAX_LIMIT
            ):
                logger.info(
                    "node({nid}) rerun times exceed max limit: {limit}".format(
                        nid=current_node.id, limit=RERUN_MAX_LIMIT
                    )
                )

                # fail
                action = Status.objects.fail(
                    current_node, "rerun times exceed max limit: {limit}".format(limit=RERUN_MAX_LIMIT)
                )

                if not action.result:
                    logger.warning(
                        "can not transit node({}) to running, pipeline({}) turn to sleep. "
                        "message: {}".format(current_node.id, process.root_pipeline.id, action.message)
                    )

                process.sleep(adjust_status=True)
                return

            if not action.result:
                logger.warning(
                    "can not transit node({}) to running, pipeline({}) turn to sleep. message: {}".format(
                        current_node.id, process.root_pipeline.id, action.message
                    )
                )
                process.sleep(adjust_status=True)
                return

            # refresh current node
            process.refresh_current_node(current_node.id)

            # build relationship
            NodeRelationship.objects.build_relationship(process.top_pipeline.id, current_node.id)
            # set up context
            context.set_node_id(current_node.id)

            result = HandlersFactory.handlers_for(current_node)(process, current_node, action.extra)

            if result.should_return or result.should_sleep:
                if result.should_sleep:
                    process.sleep(adjust_status=True)
                    if result.after_sleep_call:
                        result.after_sleep_call(*result.args, **result.kwargs)
                return

            # store current node id
            process.current_node_id = result.next_node.id



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from ast import NodeVisitor

from mako import parsetree

from pipeline.utils.mako_utils.code_extract import MakoNodeCodeExtractor
from pipeline.utils.mako_utils.exceptions import ForbiddenMakoTemplateException


class SingleLineNodeVisitor(NodeVisitor):
    """
    遍历语法树节点，遇到魔术方法使用或 import 时，抛出异常
    """

    def __init__(self, *args, **kwargs):
        super(SingleLineNodeVisitor, self).__init__(*args, **kwargs)

    def visit_Attribute(self, node):
        if node.attr.startswith("__"):
            raise ForbiddenMakoTemplateException("can not access private attribute")

    def visit_Name(self, node):
        if node.id.startswith("__"):
            raise ForbiddenMakoTemplateException("can not access private method")

    def visit_Import(self, node):
        raise ForbiddenMakoTemplateException("can not use import statement")

    def visit_ImportFrom(self, node):
        self.visit_Import(node)


class SingleLinCodeExtractor(MakoNodeCodeExtractor):
    def extract(self, node):
        if isinstance(node, parsetree.Code) or isinstance(node, parsetree.Expression):
            return node.text
        elif isinstance(node, parsetree.Text):
            return None
        else:
            raise ForbiddenMakoTemplateException("Unsupported node: [{}]".format(node.__class__.__name__))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import abc
from collections import Mapping


class DataItem(object, metaclass=abc.ABCMeta):
    def __init__(self, name, key, type, schema=None):
        self.name = name
        self.key = key
        self.type = type
        self.schema = schema

    def as_dict(self):
        return {
            "name": self.name,
            "key": self.key,
            "type": self.type,
            "schema": self.schema.as_dict() if self.schema else {},
        }


class InputItem(DataItem):
    def __init__(self, required=True, *args, **kwargs):
        self.required = required
        super(InputItem, self).__init__(*args, **kwargs)

    def as_dict(self):
        base = super(InputItem, self).as_dict()
        base["required"] = self.required
        return base


class OutputItem(DataItem):
    pass


class ItemSchema(object, metaclass=abc.ABCMeta):
    def __init__(self, description, enum=None):
        self.type = self._type()
        self.description = description
        self.enum = enum or []

    def as_dict(self):
        return {"type": self.type, "description": self.description, "enum": self.enum}

    @abc.abstractmethod
    def _type(self):
        raise NotImplementedError()


class SimpleItemSchema(ItemSchema, metaclass=abc.ABCMeta):
    pass


class IntItemSchema(SimpleItemSchema):
    @classmethod
    def _type(cls):
        return "int"


class StringItemSchema(SimpleItemSchema):
    @classmethod
    def _type(cls):
        return "string"


class FloatItemSchema(SimpleItemSchema):
    @classmethod
    def _type(cls):
        return "float"


class BooleanItemSchema(SimpleItemSchema):
    @classmethod
    def _type(cls):
        return "boolean"


class ArrayItemSchema(ItemSchema):
    def __init__(self, item_schema, *args, **kwargs):
        if not isinstance(item_schema, ItemSchema):
            raise TypeError("item_schema of ArrayItemSchema must be subclass of ItemSchema")
        self.item_schema = item_schema
        super(ArrayItemSchema, self).__init__(*args, **kwargs)

    def as_dict(self):
        base = super(ArrayItemSchema, self).as_dict()
        base["items"] = self.item_schema.as_dict()
        return base

    @classmethod
    def _type(cls):
        return "array"


class ObjectItemSchema(ItemSchema):
    def __init__(self, property_schemas, *args, **kwargs):
        if not isinstance(property_schemas, Mapping):
            raise TypeError("property_schemas of ObjectItemSchema must be Mapping type")

        if not all([isinstance(value, ItemSchema) for value in list(property_schemas.values())]):
            raise TypeError("value in property_schemas of ObjectItemSchema must be subclass of ItemSchema")

        self.property_schemas = property_schemas
        super(ObjectItemSchema, self).__init__(*args, **kwargs)

    def as_dict(self):
        base = super(ObjectItemSchema, self).as_dict()
        properties = {prop: schema.as_dict() for prop, schema in list(self.property_schemas.items())}
        base["properties"] = properties
        return base

    @classmethod
    def _type(cls):
        return "object"
